!	ArrowLISP -- An interpreter for purely symbolic LISP
!	Copyright (C) 1998-2005 Nils M Holm
!	Based upon my own implementation in T3 from 1998,
!	which was inspired by my own implementation in C from 1993.
!
!	2005-04-29	Extended syntax check of LAMBDA to cover third
!			argument.
!			Dropped dynamic scoping; removed :X meta command.
!	2005-04-28	The interpreter now makes a copy of the global
!			environment which is restored in case of an
!			error.
!	2005-04-25	Protected COPYBINDINGS() and RESTOREBINDINGS()
!			against IMPLODE and EXPLODE.
!			Made (X.NIL) => (X.NIL) rather than => (X).
!	2005-04-24	Made CALL/CC fully Scheme-compatible. Finally.
!	2005-04-21	Slightly optimized BUILTIN().
!	2005-04-19	Made EVAL tail-recursive.
!	2005-04-17	Unified MBETA and MLABL states since unbinding
!			arguments of LAMBDA and LABEL is the same.
!			Reverted DOS pool size to 12K nodes, just to
!			be on the safe side.
!	2005-04-16	Added :R meta command to turn recursive lambda
!			bindings on/off. With :R ON, LAMBDA is equal
!			to LABEL. With :R OFF, LAMBDA cannot be used to
!			bind recursive arguments.
!	2005-04-14	Added additional error checking to RESUME().
!	2005-03-27	Reduced pool size on DOS by 336 nodes due
!			to stack overflow.
!	2005-03-26	Implemented the :K (print closures) meta
!			command. Optimized FIXCLOSURES().
!	2005-03-25	Fixed lexical scoping in embedded LAMBDAs:
!			((CAR ((LAMBDA (X) (LIST (LAMBDA () X))) 'FOO)))
!			gave NIL, should have been 'FOO.
!	2005-03-21	Created a flatcopy of the B-stack when
!			resuming a captured continuation to avoid
!			an undesired side effect.
!			Removed Stack0 from continuation, since
!			it is only used in DEFINE and EVAL anyway.
!	2005-03-20	Located a bug that causes CALL/CC to fail
!			when causing a side effect in a LABEL.
!			Moved namelist of LAMBDA/LABEL to B-stack,
!			thereby fixing an interference between
!			BINDARGS() and CALL/CC.
!	2005-03-19	Documented SETUPCOND and friends.
!			Documented SETUPLABEL and friends.
!			Removed obsolete DOLABEL and DOCOND.
!	2005-03-18	Made COND evaluate its predicates in EVAL
!			so that CALL/CC can capture such predicates.
!	2005-03-17	Made LABEL bind its symbols in EVAL so that
!			CALL/CC can capture values of LABEL.
!			Added Bstack to improve performance of new
!			implementation of LABEL.
!			Made CALL/CC capture the Bstack, too.
!	2005-03-16	By removing the I-Stack, another 2K nodes
!			could be squeezed out for the DOS version.
!	2005-03-14	Made the term T of LABEL evaluate in EVAL,
!			so that CALL/CC can be used in T.
!			Fixed tail call optimization (was broken by
!			the introduction of CALL/CC).
!			Added an error message in case CALL/CC is
!			about to break the interpreter. FIXME!
!	2005-03-13	Fixed creation of closures in LABEL.
!	2005-03-12	Fixed a node leak in the CAPTURE routine.
!	2005-03-11	Rewrote EVAL so that all mutable state is
!			stored on the Lstack. Mutable state has to
!			be copied when capturing a continuation. When
!			placing mutable state on Lstack exclusively,
!			the other stacks do not have to be copied,
!			which saves both time and space.
!	2005-03-10	Started the implementation of CALL/CC. Added
!			CALL/CC and CONTINUATION keywords, where
!			(CONTINUATION ...) is passed by CALL/CC.
!			Continuations are cyclic and hence unprintable.
!			Print() takes care of this.
!			CALL/CC passed some *very* basic tests. Massive
!			testing is still to be done.
!	2005-03-09	Replaced the iteration stack (Istack) with
!			a mode stack (Mstack) and list stack (Lstack).
!			This makes state allocation during recursion
!			dynamic. The mode/list stack length is no
!			longer fixed.
!			Drawback: the interpreter does more alloc()s
!			and hence gets even slower.
!			Advantage: the entire interpreter state is
!			now stored in the node pool, thereby allowing
!			to implement CALL/CC.
!	2005-03-06	Added version tag to image header.
!			Made LAMBDA check its lexical environment.
!			Bumped major version to #12.
!	2005-01-24	Added the :X meta command to turn lexical
!			scoping on/off. If lexical scoping is off,
!			dynamic scoping takes effect. :X OFF also
!			disables the third argument of LAMBDA.
!	2004-09-24	Made LABEL create a closure when a LAMBDA
!			expression escapes its context.
!			Bumped major version to #11.
!	2004-09-23	Added optional third argument to lambda:
!			(LAMBDA (ARGS) TERM (LEXICAL-ENV))
!			where LEXICAL-ENV is an alist with bindings
!			that take effect during eval(TERM).
!			Made EVAL create a LEXICAL-ENV when a LAMBDA
!			expression escapes a local context. The
!			LEXICAL-ENV contains the bindings of all free
!			non-keyword symbols of said LAMBDA expression.
!			The effect of this change is to implement
!			lexical scoping.
!			Bumped major version to #10.
!	2004-03-06	Changed order of tests in BUILTIN().
!	2004-01-23	Removed bogous '*'s from trace output when
!			wrap column is zero.
!			Set TEXTLEN to 1024 for large environments.
!	2004-01-07	Trace output could duplicate function names.
!			Fixed.
!	2004-01-01	Fixed a small bug in the dump loader.
!			Adjusted years in copyright messages.
!	2003-12-31	Reduced large pool size to 128K nodes and
!			large iteration stack length to 32K cells.
!	2003-12-29	Made the interpreter fold its input to lower
!			case rather than upper case.
!			Made the interpreter print *any* occurence of
!			(QUOTE X) as 'X (not only at the outermost
!			level of nested S-exprs).
!			Changed default wrap column to 0.
!			Bumped major version to #9.
!	2003-12-05	Fix: The PRINTXSYM function could crash when
!			testing dotted pairs.
!	2003-12-02	Made DEFINE and DEFUN return the symbol bound
!			by them rather then *T*.
!	2003-11-23	Fix: symbols were not upcased properly in :C,
!			resulting in load buffer overflow.
!			Increased number of load buffers from 8 to 12.
!	2003-11-20	Added function call trace to error messages.
!			Call trace output is enabled and disabled using
!			the :F meta command.
!	2003-10-29	Changed error message format to:
!			>>> ERROR: FILE NAME[LINE]: FUNCTION: MESSAGE
!	2003-10-28	Implemented :W 0 (wrap mode off).
!	2003-09-30	Made :D save the :F, :I, :S, :W settings, too.
!	2003-09-21	Added support for initial load images and the the
!			:D meta command for creating node pool dumps.
!	2003-09-16	Added the special symbol ** which is bound to the
!			normal form of the most recently entered expression.
!	2003-09-12	Argument count was not checked in ATOM. Fixed.
!	2003-09-08	Added auto-implode (:I) switch. When this switch
!			is off, '#XYZ prints as '(X Y Z).
!	2003-09-07	Prefixed fatal error messages with ">>>".
!	2003-08-30	Fixed a bug in the numeric output of reduction
!			counts (printed too many leading zeroes in 1000s).
!	2003-08-29	Evaluation of symbols now costs one reduction step.
!			Made reduction counts print with commas separating
!			the 1000s.
!			Removed the stack mark routine and the -S option.
!	2003-08-14	Made the interpreter preserve line numbers when
!			loading program files.
!	2003-08-04	Made COND accept NIL as a truth value to improve
!			compatibility to other LISPs.
!			Added the *SYMBOLS* symbol which is bound to the
!			global symbol table.
!			Removed the :E meta command.
!			Added the EVAL primitive.
!	2003-08-02	Fixed quotation of #-expressions.
!			Fixed a bug that caused (NIL NIL) to crash the
!			interpreter.
!	2003-08-01	Updated meta command usage message.
!	2003-07-31	Implemented # prefix: #ABC = '(A B C).
!			Implemented auto-implosion of #-expressions.
!	2003-07-30	Implemented the IMPLODE and EXPLODE primitives.
!			Bumped major version to #8.
!	2002-09-09	Ported code to T3X 7.0.
!	2002-07-08	Made :C NAME load NAME, if NAME=>NIL. (In earlier
!			versions, only undefined NAMEs were loaded.)
!	2002-06-11	Made (LABEL ((X X)) Y) bind X to eval[X] rather
!			than (QUOTE X). This bug made
!			(LABEL ((X 'Y)) (LABEL ((X X)) X)) fail.
!			See also 2002-06-02.
!	2002-06-08	Surpressed output of => in batch mode.
!	2002-06-07	Verified that LABEL cannot be made properly
!			tail-recursive without lexical scoping.
!	2002-06-06	Fixed reduction counts: QUOTE: 0, LIST: 1 per
!			argument, but at least 1.
!	2002-06-02	Made (DEFINE X X) bind X to eval[X] rather
!			than (QUOTE X). (This was a left-over property
!			of ALISP5.)
!	2002-05-21	Removed pair flag from atoms (Cdr[x]|AFLAG) so
!			that '(X) and (X.NIL) are prefectly equal.
!			Made the pair notation (X.Y) more transparent.
!	2002-05-18	Fixed checking of maximum load level (MAXLOAD
!			is now unreachable).
!	2002-05-16	Implemented function tracking. This allows to
!			include the names of offending functions in
!			error messages. Implemented the :F meta command
!			to enable/disable function tracking.
!	2002-05-15	Added the :C (conditional load) meta command.
!	2002-05-14	The conditions of use are always printed with
!			Wrapcol=79 now.
!			Revisited some outdated comments.
!	2002-05-13	Intercepted improper argument lists in DOLAMBDA().
!	2002-05-09	Added => symbol before output of normal forms.
!			Made PRINT() quote symbols and lists.
!			Applied some minor changes.
!			Added BOTTOM function for representation of
!			undefined results.
!	2002-05-08	Moving to major version #7.
!			Removed the DEFVAR keyword. ALISP#6 compatibility
!			is hereby broken.
!			Added LABEL pseudo-function.
!			Made EVAL clean up the stack in case of an error.
!			Changed :S meta command to :S ON/OFF.
!			ERRFLAG is now reset after errors in meta commands
!			(this caused some messages to be swallowed).
!			Applied some minor cosmetics.
!	2002-02-26	Replaced the DEFVAR keyword by DEFINE.
!			Made DEFVAR an alias of DEFINE.
!	2001-02-10	Made the stack collector descend into CAR intead
!			of CDR. Thereby, the stack length could be reduced
!			from 4096 to 256 cells and still make the collector
!			more robust (since long lists are more common than
!			deep lists). The same amount of nodes is now
!			available regardless of the used collector.
!	2001-02-09	More cosmetics (replaced length() with str.length()
!			and copy() with t.memcopy()).
!	2001-02-08	Minor cosmetics and cleanup.
!	2001-02-07	Added modification date to :V (version) output.
!	2001-02-07	Now using the MFLAG constant instead of inverting
!			the bits of a visited node.
!			Notice: to compile the MFLAG constant (0x8000),
!			T3X 6.8 or later is required.
!	2001-02-07	Implemented a Deutsch/Schorr/Waite garbage
!			collector (which runs in constant space).
!			Use -S option to activate the old collector.
!	2000-09-24	Added reduction counter and :S meta command.
!	2000-09-15	Increased the number of indirect buffers from
!			three to seven.
!	2000-09-13	Replaced T with *T* and F with *F*.
!			Bound *T* to T.
!	2000-09-04	Added command line options.
!			Added batch mode.
!			Made trace output easier to process (each line
!			prefixed with '*' now).
!	2000-08-26	Added alternative QUOTE syntax: 'X = (QUOTE X).
!			Added LIST primitive.
!	2000-08-26	Translated program to T3X-R6. Reading input in
!			packed format now.
!	2000-06-16	Changed some vector sizes to make the program
!			standard compliant again.
!	2000-05-03	Misc. small cosmetical changes and fixes.
!	1999-05-02	Redirected all output through the buffered PR()
!			procedure which allows to set a wrap column (see
!			:W meta command). For use on an HP1000CX palmtop
!			computer in zoom mode.
!	1999-04-23	Implemented detection of tail-recursion in
!			EVAL(). The interpreter is now able to evaluate
!			tail-recursive expressions in constant space.
!			This is done by removing the caller's context
!			before applying a continuation.
!	1999-04-22	Replaced the BETA() function with BINDARGS()
!			and UNBINDARGS(). The evaluation step of the
!			beta reduction is now performed as a continuation
!			in EVAL().
!	1999-04-21	Made DOCOND() return its result unevaluated.
!			The evaluation of the selected conditional
!			expression is performed in EVAL() now, thereby
!			removing the (no longer needed) context of COND.
!			--
!			Recoded the MARK() routine so that is uses an
!			explicit stack instead of the system stack.
!	1999-04-17	Removed all direct recursion from EVAL() so that
!			lists are evaluated iteratively now. Indirect
!			recursion through DOCOND() and BETA() is still
!			to do.
!	1999-04-15	Changed node pool organization from a single
!			array to two arrays where one contains the CAR
!			and one the CDR fields.
!	1998-05-25	Added version number in title (currently 4).
!			Packed all message strings to squeeze out
!			more nodes for the DOS version.
!	1998-05-08	Initial implementation in T3.

decl prznum(2), pr(1), nl(0), fatal(1);

class counter()

 var	Nred, Nr1k, Nr1m, Nr1g;

 ! Reset counter.
 public reset() do
	Nred := 0;
	Nr1k := 0;
	Nr1m := 0;
	Nr1g := 0;
 end

 ! Increment counter.
 public count(k) do
	Nred := Nred+k;
	if (Nred >= 1000) do
		Nred := Nred - 1000;
		Nr1k := Nr1k + 1;
		if (Nr1k >= 1000) do
			Nr1k := 0;
			Nr1m := Nr1m+1;
			if (Nr1m >= 1000) do
				Nr1m := 0;
				Nr1g := Nr1g+1;
				if (Nr1g >= 1000)
					fatal("statistics: counter overflow");
			end
		end
	end
 end

 ! Print counter value.
 public printvalue(s) do
	if (Nr1g) do prznum(Nr1g, 0); pr(","); end
	if (Nr1m) do prznum(Nr1m, Nr1g->3:0); pr(","); end
	if (Nr1k) do prznum(Nr1k, Nr1m->3:0); pr(","); end
	prznum(Nred, Nr1k->3:0);
	pr("\s");
	pr(s);
	nl();
 end

end

module alisp(t3x, string, counter);

object	t[t3x], str[string];

const	MAJORVERSION	= 13;
var	RELEASE;

const	DEBUG	=	0;	! Debug flag
const	NBASEREG =	9;	! Number of Base Registers
const	MAXWRAPCOL =	255;	! Max. wrap column
const	MAXLOADL =	12;	! Max. load level

! Maximum length for symbols and imploded lists
const	TEXTLEN =	128;	! SMALL-ENVIRONMENT
!const	TEXTLEN =	1024;	! LARGE-ENVIRONMENT

! Number of nodes.
! (Recommended: 16-bit: 12280, >=32-bit: 131072)
const	POOLSIZE =	12280;	! SMALL-ENVIRONMENT
!const	POOLSIZE =	131072;	! LARGE-ENVIRONMENT

!	The NODE structure has a CAR and a CDR field.
!	The use of each field is as follows:
!	+---+---+--------------------------+
!	| M | A |   14-bit Address Field   |
!	+---+---+--------------------------+
!	The sign bit M is used to tag used nodes during
!	the mark step of the garbage collection.
!	The A field is used to tag atoms: if A=1 in a
!	CAR field, the node containing the field represents
!	a character (used to name atoms). In this case,
!	the CAR field contains the character itself in the
!	LSB and A=1 in the MSB. CDR points to the next
!	character or NIL. A is unused in CDR fields.
!	The node pool is kept in two separate vectors
!	called 'Car' and 'Cdr'. Consequently, Car[i] is
!	used to access the CAR field of the i'th node in
!	the pool and Cdr[i] is used to access its CDR
!	field.
!	The pointer reversing collector uses CDRs M flag
!	to store the state of an already visited atom.

! Masks for the 'A' and 'M' flags as desribed above.
const	AFLAG = 0x4000,		! SMALL-ENVIRONMENT
	MFLAG = 0x8000;		! SMALL-ENVIRONMENT
!const	AFLAG = 0x40000000,	! LARGE-ENVIRONMENT
!	MFLAG = 0x80000000;	! LARGE-ENVIRONMENT

const	EOF	=	%1;	! EOF indicator
const	DOT	=	%2;	! Internal: dot character
const	RPAREN	=	%3;	! Internal: right parenthesis

var	NIL;			! Not In List (or Pool)
var	Car[POOLSIZE],		! Vector holding CAR fields
	Cdr[POOLSIZE];		! Vector holding CDR fields
var	Infile;			! Input file name
var	Buffer,			! Ptr to current buffer
	Bufvec::MAXLOADL*TEXTLEN+MAXLOADL,	! Load buffers
	Bp, Ep,			! Input pointer, limit
	Input;			! Current input stream
var	Rejected;		! Unread character
var	Mstack, Lstack;		! Mode stack, List stack
var	Bstack;			! Binding stack, used by LABEL
var	Estack;			! Environment stack, for fixing closures
var	Parent;			! Parent pointer used in D/S/W GC
var	Free;			! Freelist
var	Symbols;		! Symbol table
var	Safesymbols;		! Safe copy of symbols
var	Stack, Stack0;		! Global stack, bottom of Stack
var	Frame;			! Current call frame
var	Tmp, Tmp2;		! Safe locations
var	Base[NBASEREG];		! Base registers (roots for GC)
var	Level;			! Nesting level during input
var	Loadlev;		! Nesting level of LOAD meta cmds
var	Evlev;			! Number of nested EVALs
var	Errflag;		! Error flag
var	Function;		! Name of current lambda function
var	Trace;			! Name of traced function
var	Prtbuf::MAXWRAPCOL+1,	! Print buffer,
	Pp;			! Pointer
var	Quoted;			! Quote flag of PRINT
var	Nlseq::3, Nll;		! Newline sequence, length
var	Wrapcol;		! Wrap column
var	Prefix;			! Output line prefix
var	Maxuse, Lastuse;	! Memory use gauge
var	Line;			! Input line number
object	Reductions[counter],	! Reduction counter
	Allocations[counter];	! Allocation counter
var	Statflag;		! Statistics flag
var	Ftflag;			! Function tracking flag
var	Ntrace;			! Max functions to print in call trace
var	Implflag;		! Auto-implode flag
var	Prclflag;		! Print closures flag
var	Lexenv;			! Environment for creating closures
var	Reclflag;		! Recursive lambda bindings flag
var	Bound;			! Variables bound in a closure
var	Image::TEXTLEN;		! Name of initial load image

! Builtin symbol pointers (for fast lookup)
var	S_atom, S_bottom, S_callcc, S_car, S_cdr, S_cond, S_cons,
	S_continuation, S_define, S_defun, S_eq, S_eval, S_explode,
	S_false, S_implode, S_label, S_lambda, S_list, S_nil, S_null,
	S_quote, S_symbols, S_t, S_true, S_last;

! Command line flags
var	O_batch;	! Batch (non-interactive) mode

! Emit a newline sequence
newline() t.write(T3X.SYSOUT, Nlseq, Nll);

! Print the string S thru a buffered interface.
pr(s) do var k, n, lim;
	if (\s) do
		if (Prefix /\ Wrapcol)
			t.write(T3X.SYSOUT, Prefix, str.length(Prefix));
		t.write(T3X.SYSOUT, PrtBuf, Pp);
		Pp := 0;
		return 0;
	end
	k := str.length(s);
	lim := Wrapcol-> Wrapcol: MAXWRAPCOL;
	while (Pp + k > lim) do
		if (Prefix /\ Wrapcol)
			t.write(T3X.SYSOUT, Prefix, str.length(Prefix));
		n := lim - Pp;
		t.memcopy(@Prtbuf::Pp, s, n);
		t.write(T3X.SYSOUT, Prtbuf, lim);
		Pp := 0;
		k := k-n;
		s := @s::n;
		if (Wrapcol) newline();
	end
	if (k) do
		t.memcopy(@Prtbuf::Pp, s, k);
		Pp := Pp+k;
	end
end

! Print a padded number with leading characters.
_prnum(n, w, spaces) do var b::20, k;
	k := str.length(str.numtostr(b, n, 10));
	if (k < w) pr(@spaces::(k));
	pr(b);
end

! Print number with leading spaces.
prnum(n, w) _prnum(n, w, "\s\s\s\s\s\s");

! Print number with leading zeroes.
prznum(n, w) _prnum(n, w, "000");

! Flush the output buffer and emit a newline sequence.
nl() do
	pr(0);
	newline();
end

! Print message M and halt the interpreter.
fatal(m) do
	t.write(T3X.SYSOUT, ">>> Fatal: ", 11);
	t.write(T3X.SYSOUT, m, str.length(m));
	t.write(T3X.SYSOUT, Nlseq, Nll);
	halt 1;
end

! Convert string to lower case.
locase(s) do var k, i;
	k := str.length(s)+1;
	for (i=0, k)
		if ('A' <= s::i /\ s::i <= 'Z')
			s::i := s::i + ('a' - 'A');
	return s;
end

decl print(1), _print(1);

! Print function names on call stack
tracecall() do var s, v, n;
	s := Frame;
	n := Ntrace;
	while (s \= NIL) do
		if (\n \/ Cdr[s] = NIL \/ Car[Cdr[s]] = NIL) leave;
		if (n = Ntrace) pr(">>> Trace:");
		n := n-1;
		pr(" ");
		Quoted := 1;
		_print(Car[Cdr[s]]);
		s := Car[s];
	end
	if (n \= Ntrace) nl();
end

! Print error message M and set Errflag.
! If N \= %1 print an additional line containing '>>> N'.
error(m, n) do
	if (Errflag) return NIL;
	pr(">>> Error: ");
	if (Infile) do
		pr(locase(Infile));
	end
	pr("[");
	prnum(Line, 0);
	pr("]: ");
	ie (\Ftflag) do
		pr("(unknown)");
	end
	else ie (Function \= NIL) do
		Quoted := 1;
		_print(Function);
	end
	else do
		pr("(top level)");
	end
	pr(": ");
	pr(m);
	nl();
	if (n \= %1) do
		pr(">>> "); print(n); nl();
	end
	if (Ftflag /\ Frame \= NIL) tracecall();
	if (O_batch) halt 1;
	Errflag := %1;
	Function := NIL;
	return NIL;
end

! Mark nodes which can be accessed through N.
! This routine uses the Deutsch/Schorr/Waite algorithm
! (aka pointer reversal algorithm) which marks the
! nodes of a pool in constant space.
! In addition to the Mark flag (MFLAG), it uses
! another bit to save the state of the current
! node (S). The (otherwise unused) MFLAG of CDR is
! used to store the state.
! Each visited node goes through these states:
! M=0 S=0 unvisited, process CAR
! M=1 S=1 CAR visited, process CDR
! M=1 S=0 completely visited, return to parent
mark(n) do var p;
	Parent := NIL;	! Initially, there is no parent node
	while (1) do
		! Reached a leaf?
		ie (n = NIL \/ Car[n] & MFLAG) do
			! When the current node is a leaf and there is
			! no parent, the entire tree is marked.
			if (Parent = NIL) leave;
			ie (Cdr[Parent] & MFLAG) do
				! State 2: the CDR of the parent has
				! not yet been marked (M of CDR set).
				! Swap CAR and CDR pointers and
				! proceed with CDR. Set State=3.
				p := Cdr[Parent] & ~MFLAG;
				Cdr[Parent] := Car[Parent] & ~MFLAG;	! S=0
				Car[Parent] := n | MFLAG;		! M=1
				n := p;
			end
			else do
				! State 3: CAR and CDR of parent done.
				! Return to the parent and restore
				! parent of parent.
				p := Parent;
				Parent := Cdr[p];
				Cdr[p] := n;
				n := p;
			end
		end
		else do
			! State 1: The current node has not yet been
			! visited.
			ie (Car[n] & AFLAG) do
				! If the node is an atom, go directly
				! to state 3: Save the parent in CDR,
				! make the current node the new parent
				! and move to its CDR.
				p := Cdr[n];
				Cdr[n] := Parent;			! S=0
				Parent := n;
				n := p;
				Car[Parent] := Car[Parent] | MFLAG;	! M=1
			end
			else do
				! Go to state 2: like above, but save
				! the parent in CAR and proceed to CAR.
				p := Car[n];
				Car[n] := Parent | MFLAG;		! M=1
				Parent := n;
				n := p;
				Cdr[Parent] := Cdr[Parent] | MFLAG;	! S=1
			end
		end
	end
end

! Mark and Sweep Garbage Collection.
! First, mark all nodes which can be accessed through
! Base Registers (Base[]) and then reclaim untagged
! nodes.
gc() do var i, n, k;
	k := 0;
	if (DEBUG) do pr("GC called"); nl(); end
	for (i=0, NBASEREG) mark(Base[i][0]);
	Free := NIL;
	for (i=0, POOLSIZE) do
		ie (\(Car[i] & MFLAG)) do
			Cdr[i] := Free;
			Free := i;
			k := k+1;
		end
		else do
			Car[i] := Car[i] & ~MFLAG;
		end
	end
	if (Maxuse < POOLSIZE-k) Maxuse := POOLSIZE-k;
	if (DEBUG) do
		prnum(k, 0); pr(" nodes reclaimed"); nl();
	end
	return k;
end

! Allocate a fresh node and initialize with PCAR,PCDR.
alloc(pcar, pcdr) do var n;
	if (Statflag) Allocations.count(1);
	if (Free = NIL) do
		gc();
		if (Free = NIL) fatal("out of nodes");
	end
	n := Free;
	Free := Cdr[Free];
	Car[n] := pcar;
	Cdr[n] := pcdr;
	return n;
end

! Save node N on the Stack.
save(n) do
	Tmp := n;	! Otherwise, alloc() might recycle this node.
	Stack := alloc(n, Stack);
	Tmp := NIL;
end

! Pop K nodes off the Stack and return
! the most recently popped one.
unsave(k) do var n;
	while (k) do
		if (Stack = NIL) fatal("stack underflow");
		n := Car[Stack];
		Stack := Cdr[Stack];
		k := k-1;
	end
	return n;
end

! Save value V on the M-Stack.
msave(v) do
	! Since the Mstack holds integer values rather than
	! nodes, the values are packaged in the character
	! fields of atoms.
	Car[Mstack] := alloc(v|AFLAG, Car[Mstack]);
end

! Pop a value off the M-Stack and return it.
munsave() do var v;
	if (Car[Mstack] = NIL) fatal("m-stack underflow");
	v := Car[Car[Mstack]] & ~AFLAG;		! See msave()
	Car[Mstack] := Cdr[Car[Mstack]];
	return v;
end

! Save node N on the L-Stack.
lsave(n) do
	Tmp := n;	! Otherwise, alloc() might recycle this node.
	Lstack := alloc(n, Lstack);
	Tmp := NIL;
end

! Pop K nodes off the L-Stack and return
! the most recently popped one.
lunsave(k) do var n;
	while (k) do
		if (Lstack = NIL) fatal("l-stack underflow");
		n := Car[Lstack];
		Lstack := Cdr[Lstack];
		k := k-1;
	end
	return n;
end

! Save node N on the B-Stack.
bsave(n) do
	Tmp := n;	! Otherwise, alloc() might recycle this node.
	Bstack := alloc(n, Bstack);
	Tmp := NIL;
end

! Pop K nodes off the B-Stack and return
! the most recently popped one.
bunsave(k) do var n;
	while (k) do
		if (Bstack = NIL) fatal("b-stack underflow");
		n := Car[Bstack];
		Bstack := Cdr[Bstack];
		k := k-1;
	end
	return n;
end

! Read a single character from the input stream
! and return it. Rdch()=EOF indicates that the
! input is exhausted.
! Letters will be automatically converted into
! upper case and stripped to seven bits.
rdch() do var c;
	if (Rejected \= EOF) do
		c := Rejected;
		Rejected := EOF;
		return c;
	end
	if (Bp >= Ep) do
		Ep := t.read(Input, Buffer, TEXTLEN);
		Bp := 0;
	end
	if (Ep < 1) return EOF;
	Bp := Bp+1;
	c := Buffer::(Bp-1) & 127;
	if (c = '\n') Line := Line+1;
	return ('A' <= c /\ c <= 'Z')-> c-'A'+'a' : c;
end

! Find a symbol named S in the symbol table.
! Each symbol is represented by a (NAME.VALUE) pair
! where NAME is a list of character nodes and value
! may be any valid S-expression.
! The symbol table is a list containing symbol
! pairs ((N1.V1) ...).
! When a symbol named S is found, return its
! pair (S.V) and otherwise return NIL.
findsym(s) do var y, n, i;
	y := Symbols;
	while (y \= NIL) do
		n := Car[Car[y]];
		i := 0;
		while (n \= NIL /\ s::i) do
			if (s::i \= Car[n] & 255) leave;
			n := Cdr[n];
			i := i+1;
		end
		if (n = NIL /\ \s::i) return Car[y];
		y := Cdr[y];
	end
	return NIL;
end

! Add the symbol S to the symbol table if it
! does not already exist. If it does exist,
! return the existing symbol.
! When adding a new symbol, initialize the
! VALUE field with V. If V=0, make the symbol
! a constant (bind it to itself).
! Return the pair representing the symbol S.
addsym(s, v) do var i, n, m, a;
	i := 0;
	n := findsym(s);
	if (n \= NIL) return n;
	while (s::i) do
		m := alloc(s::i | AFLAG, NIL);
		ie (n = NIL) do		! Protect the first character
			n := m;
			save(n);
		end
		else do		! Just append the rest
			Cdr[a] := m;
		end
		a := m;
		i := i+1;
	end
	m := alloc(n, v-> v: n);
	save(m);
	Symbols := alloc(m, Symbols);
	if (S_symbols \= NIL) Cdr[S_symbols] := Symbols;
	unsave(2);
	return m;
end

decl	read(0);

! Read a list (S0 ... SN) and return (a pointer to) it.
! This routine also recognizes pairs of the form (S0.S1).
! For empty lists, it returns NIL.
rdlist() do
	var	n,	! Node read
		m,	! Ptr to list
		a,	! Used to append nodes to m
		c;	! Member counter
	var	badpair;

	badpair := "bad pair";
	Level := Level+1;
	m := alloc(NIL, NIL);	! Root node
	save(m);
	a := NIL;
	c := 0;
	while (1) do
		n := read();
		if (n = EOF) fatal("EOF in list");
		if (n = DOT) do
			if (c \= 1) do
				error(badpair, %1);
				loop;
			end
			Cdr[m] := read();
			if (read() \= RPAREN) do
				error(badpair, %1);
				loop;
			end
			unsave(1);
			Level := Level-1;
			return m;
		end
		if (n = RPAREN) leave;
		ie (a = NIL) 
			a := m;		! First member: insert at root
		else
			a := Cdr[a];	! Following members: append
		Car[a] := n;
		Cdr[a] := alloc(NIL, NIL);	! Alloc space for next member
		c := c+1;
	end
	Level := Level-1;
	unsave(1);
	if (a \= NIL) Cdr[a] := NIL;	! Remove trailing empty node
	return c-> m: NIL;
end

! Read characters from the input device until a non-space
! character [^\t\s] is found. Return that character.
skip() do var c;
	c := rdch();
	while (c = '\s' \/ c = '\t') c := rdch();
	return c;
end

imagevars() return [
	@Ftflag, @Statflag, @Implflag, @Prclflag, @Wrapcol,
	@Reclflag, @Symbols, @Free,
	@S_atom, @S_bottom, @S_callcc, @S_car, @S_cdr, @S_cond, @S_cons,
	@S_continuation, @S_define, @S_defun, @S_eq, @S_eval, @S_explode,
	@S_false, @S_implode, @S_label, @S_lambda, @S_list, @S_nil, @S_null,
	@S_quote, @S_symbols, @S_t, @S_true, @S_last,
0 ];

! Dump node pool image to file named in P.
dump(p) do var fd, n, v, i, magic;
	fd := t.open(p, T3X.OWRITE);
	if (fd < 0) do
		error("cannot create file", %1);
		pr(">>> "); pr(p); nl();
		return NIL;
	end
	magic := "ALISP__";
	magic::5 := t.bpw();
	magic::6 := MAJORVERSION;
	t.write(fd, magic, 7);
	n := POOLSIZE;
	t.write(fd, @n, t.bpw());
	v := imagevars();
	i := 0;
	while (v[i]) do
		t.write(fd, v[i], t.bpw());
		i := i+1;
	end
	if (	t.write(fd, Car, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw() \/
		t.write(fd, Cdr, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw()
	) do
		t.close(fd);
		pr("dump failed"); nl();
		return;
	end
	t.close(fd);
	if (\O_batch) do
		pr("dumped"); nl();
	end
end

decl toploop(0);

! Load S-expressions from the external file or device
! named in the string P. Return nothing.
load(p) do
	var	o, of, buf, b, e, r;
	var	fd, oline;

	if (Loadlev+1 >= MAXLOADL) return error("load level too deep", %1);
	fd := t.open(p, T3X.OREAD);
	if (fd < 0) do
		error("cannot open file", %1);
		pr(">>> "); pr(p); nl();
		return NIL;
	end
	Loadlev := Loadlev + 1;
	! Save old I/O state
	buf := Buffer;
	b := Bp;
	e := Ep;
	r := Rejected;
	! Reset I/O buffer
	Bp := 0;
	Ep := 0;
	Rejected := EOF;
	Buffer := @Bufvec::((TEXTLEN+1)*Loadlev);
	! Run the toplevel loop with redirected I/O
	o := Input;
	Input := fd;
	oline := Line;
	Line := 1;
	of := Infile;
	Infile := p;
	toploop();
	Infile := of;
	Line := oline;
	! Restore previous I/O state
	Rejected := r;
	Ep := e;
	Bp := b;
	Buffer := buf;
	Loadlev := Loadlev - 1;
	Input := o;
	t.close(fd);
	if (Level) error("unbalanced parens in load", %1);
	if (\Loadlev /\ \O_batch) do
		pr("loaded"); nl();
	end
end

! Read all characters up to the next line terminator (or the EOF)
! from the input device and store them in the vector P.
! If L\=0, lowcase each character read.
! Return the terminating character.
! This routine is used to accept parameters to meta commands.
getname(p, l) do var c, i;
	c := skip();
	i := 0;
	while (c \= '\n' /\ c \= '\r' /\ c \= EOF) do
		! Lowcase characters to avoid
		! SUCH_NAMES under Unix.
		if (l /\ 'A' <= c /\ c <= 'Z') c := c-'A'+'a';
		p::i := c;
		if (i >= TEXTLEN-2) do
			error("name too long", %1);
			return 0;
		end
		c := rdch();
		i := i+1;
	end
	p::i := 0;
	return c;
end

! Extract a YES/NO, ON/OFF, 0/1 argument used to
! enable and disable options.
yesno(p, pf) do var c, f;
	c := getname(p, 1);
	f := 2;
	if (p::0 = 'y' \/ p::0 = '1') f := %1;
	if (p::0 = 'n' \/ p::0 = '0') f := 0;
	if (p::0 = 'o') do
		if (p::1 = 'n') f := %1;
		if (p::1 = 'f') f := 0;
	end
	if (f = 2) error("Usage: command on/off", %1);
	pf[0] := f;
	return c;
end

! Count the nodes in the freelist.
countfree() do var n, k;
	k := 0;
	n := Free;
	while (n \= NIL) do
		k := k+1;
		n := Cdr[n];
	end
	return k;
end

! Print a line followed by a newline sequence.
pl(s) do
	pr(s); nl();
end


! Print the conditions of use.
license() do var w;
	w := Wrapcol;
	Wrapcol := 79;
pl(
"ArrowLISP -- An Interpreter for Purely Symbolic LISP");
pl(
"Copyright (C) 1998-2005 Nils M Holm.  All rights reserved.");
pl("");
pl(
"Redistribution and use in source and binary forms, with or without");
pl(
"modification, are permitted provided that the following conditions");
pl(
"are met:");
pl(
"1. Redistributions of source code must retain the above copyright");
pl(
"   notice, this list of conditions and the following disclaimer.");
pl(
"2. Redistributions in binary form must reproduce the above copyright");
pl(
"   notice, this list of conditions and the following disclaimer in the");
pl(
"   documentation and/or other materials provided with the distribution.");
pl("");
pl(
"THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND");
pl(
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE");
pl(
"IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE");
pl(
"ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE");
pl(
"FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL");
pl(
"DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS");
pl(
"OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)");
pl(
"HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT");
pl(
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY");
pl(
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF");
pl(
"SUCH DAMAGE.");
pl("");
	Wrapcol := w;
end

! Normal termination
bye() do
	if (\O_batch) do
		pr("Bye"); nl();
	end
	halt;
end

! Set and acknowledge state of flag
setflag(name, varp, val) do
	if (val \= 2) do
		varp[0] := val;
		if (\O_batch) do
			pr(name);
			pl(varp[0]-> " on": " off");
		end
	end
end

! Read and process meta (colon) commands.
meta() do
	var	c, k, k2, y;
	var	p::TEXTLEN+1;

	c := skip();
	ie (c = 'c') do			! Load conditionally
		c := getname(p, 1);
		k := str.scan(p, '/')+1;
		k2 := str.scan(p, '.');
		if (k2 \= %1) p::k2 := 0;
		y := findsym(locase(@p::k));
		if (k2 \= %1) p::k2 := '.';
		ie (y = NIL \/ Cdr[y] = NIL) do
			load(locase(p));
		end
		else if (\Loadlev /\ \O_batch) do
			pr("found"); nl();
		end
	end
	else ie (c = 'd') do		! Dump node pool image
		c := getname(p, 1);
		dump(p);
	end
	else ie (c = 'f') do		! Switch statistics mode on/off
		c := yesno(p, @k);
		setflag("function tracking", @Ftflag, k);
	end
	else ie (c = 'g') do		! Trigger GC
		k := countfree();
		k2 := gc();
		if (Maxuse \= 0) Lastuse := Maxuse;
		Maxuse := 0;
		prnum(Lastuse, 6);
		pr(" nodes used max."); nl();
		prnum(k2-k, 6);
		pr(" nodes reclaimed"); nl();
		prnum(k2, 6);
		pr(" nodes in freelist"); nl();
	end
	else ie (c = 'i') do		! Switch auto-implosion on/off
		c := yesno(p, @k);
		setflag("auto-implode", @Implflag, k);
	end
	else ie (c = 'k') do		! Switch closure printing on/off
		c := yesno(p, @k);
		setflag("closure printing", @Prclflag, k);
	end
	else ie (c = 'l') do		! Load program
		c := getname(p, 1);
		load(p);
	end
	else ie (c = 'q') do		! Quit
		bye();
	end
	else ie (c = 'r') do		! Switch recursive lambda on/off
		c := yesno(p, @k);
		setflag("recursive lambda bindings", @Reclflag, k);
	end
	else ie (c = 's') do		! Switch statistics mode on/off
		c := yesno(p, @k);
		setflag("statistics", @Statflag, k);
	end
	else ie (c = 't') do		! Specify/reset function to trace
		c := getname(p, 0);
		Trace := findsym(p);
		ie (Trace \= NIL) do
			if (\O_batch) do
				pr("tracing "); pr(p); nl();
			end
		end
		else ie (p::0) do
			error("no such symbol", %1);
			pr(">>> "); pr(p); nl();
		end
		else do
			if (\O_batch) do
				pr("trace off"); nl();
			end
		end
	end
	else ie (c = 'v') do		! Version info or License
		c := getname(p, 0);
		ie (p::0) do
			license();
		end
		else do
			pr("ArrowLISP (C) 1998-2005 Nils M Holm"); nl();
			pr("Latest modification: ");
			pr(RELEASE); nl();
			prnum(POOLSIZE, 6);
			pr(" total nodes"); nl();
		end
	end
	else ie (c = 'w') do		! Specify wrap column
		c := getname(p, 0);
		k := str.strtonum(p, 10, 0);
		ie (\k) do
			Wrapcol := 0;
			if (\O_batch) do
				pr("wrap=off"); nl();
			end
		end
		else ie (k < 1 \/ k > MAXWRAPCOL) do
			error("bad wrap column", %1);
		end
		else do
			Wrapcol := k;
			if (\O_batch) do
				pr("wrap="); prnum(Wrapcol, 0); nl();
			end
		end
	end
	else do
		pr("Usage: :c file - load file conditionally"); nl();
		pr("       :d file - dump image to file"); nl();
		pr("       :f 0/1  - function tracking on/off"); nl();
		pr("       :g      - collect garbage"); nl();
		pr("       :i 0/1  - auto-implode lists on/off"); nl();
		pr("       :k 0/1  - print closures on/off"); nl();
		pr("       :l file - load file"); nl();
		pr("       :q      - quit"); nl();
		pr("       :r 0/1  - rec lambda bindings on/off"); nl();
		pr("       :s 0/1  - statistics on/off"); nl();
		pr("       :t name - trace function (:t=off)"); nl();
		pr("       :v[l]   - show version/license"); nl();
		pr("       :w col  - set wrap column (0=none)"); nl();
	end
	while (c \= '\n' /\ c \= EOF) c := rdch();
end

! Is N a 'real' (non-NIL) Atom?
atomic(n) return n \= NIL /\ Car[n] \= NIL /\ Car[Car[n]] & AFLAG;

decl explode(1);

! Read an expression from the current input stream
! and return (a pointer to) it.
! Each valid expression is either an Atom or a List.
read() do var c, l, q;
	c := rdch();
	while (1) do	! Skip spaces and comments, run meta commands
		while (c = '\s' \/ c = '\t' \/ c = '\n' \/ c = '\r')
			c := rdch();
		if (c = ':' /\ \Level) do
			meta();
			c := '\n';
			Errflag := 0;
			loop;
		end
		if (c \= ';') leave;
		while (c \= '\n') c := rdch();
	end
	if (c = EOF) return EOF;
	ie (c = '(') do
		return rdlist();
	end
	else ie (c = '\'') do
		l := read();
		save(l);
		q := alloc(S_quote, NIL);
		save(q);
		Cdr[q] := alloc(l, NIL);
		unsave(2);
		return q;
	end
	else ie (c = '#') do
		l := read();
		if (\atomic(l)) return error("non-atom after '#'", %1);
		return explode(l);
	end
	else ie (c = ')') do
		if (\Level) return error("unexpected ')'", %1);
		return RPAREN;
	end
	else ie (c = '.') do
		if (\Level) return error("unexpected '.'", %1);
		return DOT;
	end
	else do	var s::TEXTLEN, i;
		! Accept symbol and add to symbol table
		i := 0;
		while (c \= '\s' /\ c \= '\t' /\ c \= '\n' /\
			c \= '\r' /\ c \= '(' /\ c \= ')' /\
			c \= ';' /\ c \= '.' /\ c \= '#'
		) do
			if (i >= TEXTLEN-2) do
				error("symbol too long", %1);
				i := i-1;
			end
			s::i := c;
			i := i+1;
			c := rdch();
		end
		s::i := 0;
		Rejected := c;
		return addsym(s, NIL);
	end
end

! Misc. error reporting and handling routines...

wrongargs(n) return error("wrong argument count", n);

badargl(n) return error("bad argument list", n);

! Evaluate N=(CONS M M2)
docons(n) do var m, m2;
	if (Statflag) Reductions.count(1);
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	m2 := Car[Cdr[m]];
	m := alloc(Car[m], m2);
	return m;
end

! Evaluate N=(LIST M1 ... Mn)
dolist(n) do var l, x, nm;
	l := NIL;
	nm := 1;
	while (1) do
		n := Cdr[n];
		if (n = NIL) leave;
		ie (l = NIL) do
			l := alloc(Car[n], NIL);
			x := l;
			save(l);
		end
		else do
			Cdr[x] := alloc(Car[n], NIL);
			x := Cdr[x];
			nm := nm+1;
		end
	end
	if (Statflag) Reductions.count(nm);
	if (l \= NIL) unsave(1);
	return l;
end

! Evaluate N=(CAR M)
docar(n) do var m;
	if (Statflag) Reductions.count(1);
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	if (atomic(Car[m]) \/ Car[m] = NIL)
		return error("bad pair in 'car'", n);
	return Car[Car[m]];
end

! Evaluate N=(CDR M)
docdr(n) do var m;
	if (Statflag) Reductions.count(1);
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	if (atomic(Car[m]) \/ Car[m] = NIL)
		return error("bad pair in 'cdr'", n);
	m := Cdr[Car[m]];
	return m;
end

! Explode a symbol into one-character symbols
explode(n) do var l, x, nm, name::2, y;
	n := Car[n];
	l := NIL;
	nm := 1;
	name::1 := 0;
	while (1) do
		if (n = NIL) leave;
		name::0 := Car[n];
		y := addsym(name, NIL);
		ie (l = NIL) do
			l := alloc(y, NIL);
			x := l;
			save(l);
		end
		else do
			Cdr[x] := alloc(y, NIL);
			x := Cdr[x];
			nm := nm+1;
		end
		n := Cdr[n];
	end
	if (Statflag) Reductions.count(nm);
	if (l \= NIL) unsave(1);
	return l;
end

! Evaluate N=(EXPLODE M)
doexplode(n) do
	n := Cdr[n];
	if (n = NIL \/ Cdr[n] \= NIL) return wrongargs(n);
	if (Car[n] = NIL) do
		if (Statflag) Reductions.count(1);
		return NIL;
	end
	if (\atomic(Car[n])) return error("non-atom in 'explode'", Car[n]);
	return explode(Car[n]);
end

! Evaluate N=(IMPLODE M)
doimplode(n) do var name::TEXTLEN+1, i;
	n := Cdr[n];
	if (n = NIL \/ Cdr[n] \= NIL) return wrongargs(n);
	if (atomic(Car[n])) return error("non-list in 'implode'", Car[n]);
	n := Car[n];
	i := 0;
	while (1) do
		if (n = NIL) leave;
		if (\atomic(Car[n]))
			return error("non-atom in 'implode' list", Car[n]);
		if (Cdr[Car[Car[n]]] \= NIL)
			return error("symbol too long in 'implode'", Car[n]);
		name::i := Car[Car[Car[n]]];
		if (i >= TEXTLEN-1)
			return error("list too long in 'implode'", Car[n]);
		i := i+1;
		n := Cdr[n];
	end
	if (Statflag) Reductions.count(i+1);
	name::i := 0;
	return i = 0-> NIL: addsym(name, NIL);
end

! Evaluate N=(NULL M)
donull(n) do var m;
	if (Statflag) Reductions.count(1);
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	return Car[m] = NIL-> S_true: S_false;
end

! Evaluate N=(EQ M M2)
doeq(n) do var m;
	if (Statflag) Reductions.count(1);
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	return Car[m] = Car[Cdr[m]]-> S_true: S_false;
end

! Evaluate N=(ATOM M)
doatom(n) do var m;
	if (Statflag) Reductions.count(1);
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	return atomic(Car[m]) \/ Car[m] = NIL -> S_true: S_false;
end

! Extract clause from argument list of COND.
! Check the syntax of the clause.
! Return the predicate of the clause for
! evaluation in EVAL.
getpred() do var e;
	e := Car[Car[Bstack]];
	if (	atomic(e) \/ e = NIL \/
		Cdr[e] = NIL \/ Cdr[Cdr[e]] \= NIL
	)
		return error("bad clause in 'cond'", e);
	if (Statflag) Reductions.count(1);
	return Car[e];	! predicate
end

! Setup context for evaluation of (COND (P1 E1) ... (Pn En)).
! The context consits of a list of clauses.
! return the predicate of the first clause.
setupcond(n) do var m;
	m := Cdr[n];
	if (m = NIL) return wrongargs(n);
	bsave(m);
	return getpred();
end

! Evaluate next clause of COND.
! N is the value of the current predicate.
! If N=*T*, return the expression of the predicate.
! If N=*F*, return the predicate of the next clause.
! If N=NIL, see N=*F*.
! When returning the expression of a predicate (N=*T*),
! set the context on the Bstack to NIL to signal that
! a true clause was found.
! Clauses and predicates are returned to EVAL for
! evaluation.
evalclause(n) do var e;
	e := Car[Bstack];
	ie (n = S_false \/ n = NIL) do
		Car[Bstack] := Cdr[e];
		if (Car[Bstack] = NIL)
			return error("no default in 'cond'", %1);
		return getpred();
	end
	else ie (n = S_true) do
		e := Car[Cdr[Car[e]]];
		Car[Bstack] := NIL;
		return e;
	end
	else do
		return error("bad predicate in 'cond'", Car[e]);
	end
end

! Unbind the arguments of LAMBDA and LABEL.
! See also BINDARGS().
unbindargs() do var v;
	if (Ftflag) do
		Frame := unsave(1);
		Function := unsave(1);
	end
	v := bunsave(1);		! Caller's namelist
	while (v \= NIL) do
		Cdr[Car[v]] := unsave(1);
		v := Cdr[v];
	end
end

! Check if the symbol N is bound in the current lexical
! environment.
isbound(n) do var b;
	b := Bound;
	while (b \= NIL) do
		if (n = Car[b]) return 1;
		b := Cdr[b];
	end
	b := Car[Lexenv];
	while (b \= NIL) do
		if (Car[Car[b]] = n) return 1;
		b :=Cdr[b];
	end
	return 0;
end

! Recursively collect free variables and add their symbols
! and values to the current lexical environment.
collect(n) do
	if (n = NIL \/ n & AFLAG \/ Car[n] & AFLAG) return;
	if (Car[Car[n]] & AFLAG) do
		if (Car[n] = Cdr[n] \/ isbound(n)) return;
		Car[Lexenv] := alloc(NIL, Car[Lexenv]);
		Car[Car[Lexenv]] := alloc(n, Cdr[n]);
		return;
	end
	collect(Car[n]);
	collect(Cdr[n]);
end

! Create lexical environment.
mklexenv(term, locals) do var env;
	Lexenv := alloc(NIL, NIL);
	save(Lexenv);
	Bound := locals;
	collect(term);
	unsave(1);
	return Car[Lexenv];
end

! Create a closure from a lambda expression.
closure(n) do var cl, env, args, term;
	if (Errflag) return NIL;
	cl := alloc(S_lambda, NIL);
	save(cl);
	args := Car[Cdr[n]];
	Cdr[cl] := alloc(args, NIL);
	term := Car[Cdr[Cdr[n]]];
	Cdr[Cdr[cl]] := alloc(term, NIL);
	env := mklexenv(term, args);
	save(env);
	if (env \= NIL) do
		Cdr[Cdr[Cdr[cl]]] := alloc(env, NIL);
		Estack := alloc(env, Estack);
	end
	unsave(2);
	return cl;
end

! Fix recursive bindings in closures.
fixclosures() do var aa, a, ee, e;
	if (Errflag \/ Estack = NIL) return;
	a := Car[Bstack];
	while (a \= NIL) do
		ee := Estack;
		while (ee \= NIL) do
			e := Car[ee];
			while (e \= NIL) do
				if (Car[a] = Car[Car[e]]) do
					Cdr[Car[e]] := Cdr[Car[a]];
					leave;
				end
				e := Cdr[e];
			end
			ee := Cdr[ee];
		end
		a := Cdr[a];
	end
	Estack := NIL;
end

! Set up a context for processing N=(LABEL ((MA1 eval[MX2]) ...) MN)
! Save
! - the complete LABEL expression on the Bstack
! - the environment on the Bstack
! - two copies of the saved name list on the Stack
setuplabel(n) do
	var	m,	! Argument pointer
		v;	! List of re-bound names

	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	m := Car[m];
	if (atomic(m)) return error("bad environment in 'label'", m);
	bsave(n);	! save entire LABEL
	bsave(m);	! save environment
	bsave(NIL);	! save empty name list
	return m;
end

! Process one binding of a LABEL:
! bind value to name, advance to next binding.
! Return:
! non-NIL - more bindings in environment
! NIL     - last binding done
nextlabel(n) do var m, p, r;
	m := Car[Cdr[Bstack]];		! rest of environment
	if (m = NIL) return NIL;
	p := Car[m];
	Cdr[Car[p]] := n;		! Bind new value
	if (Statflag) Reductions.count(1);
	Car[Cdr[Bstack]] := Cdr[m];
	return Cdr[m];
end

! Evaluate value to bind inside a LABEL:
! - check LABEL syntax
! - save name to bind to
! - save original binding of name
! - return value for evaluation in EVAL
evallabel() do
	var	m, p, v;

	m := Car[Cdr[Bstack]];
	p := Car[m];
	! Each binding must have the form (atom expr)
	if (	atomic(p) \/ Cdr[p] = NIL \/ atomic(Cdr[p]) \/
		Cdr[Cdr[p]] \= NIL \/ \atomic(Car[p])
	) do
		! In case of an error, get rid of the
		! partial environment.
		v := bunsave(1);
		bunsave(2);
		bsave(v);
		if (Ftflag) do
			save(Function);
			save(Frame);
		end
		unbindargs();
		return error("bad binding in 'label'", p);
	end
	Car[Bstack] := alloc(Car[p], Car[Bstack]);	! Save name
	save(Cdr[Car[p]]);				! Save value
	! Evaluate the new value of the current symbol
	return Car[Cdr[p]];
end

! Finish processing bindings of LABEL:
! finish context and return term of LABEL.
finishlabel(tail) do var m, v;
	v := bunsave(1);
	m := bunsave(2);		! drop environment, get full LABEL
	bsave(v);
	fixclosures();
!	if (tail) do	! XXX to do
!		! remove caller's context
!		! (which is BELOW the current context!)
!		Tmp2 := unsave(1); ! M
!		unbindargs();
!		unsave(1);
!		munsave();
!		save(Tmp2);
!		Tmp2 := NIL;
!	end
	if (Ftflag) do			! required by UNBINDARGS()
		save(Function);
		save(Frame);
	end
	return Car[Cdr[Cdr[m]]];	! return term of LABEL
end

! Evaluate N=(DEFUN M (M2) M3)
! The name M already has been added to the
! symbol table by READ().
dodefun(n) do var m, f;
	m := Cdr[n];
	if (	m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] = NIL \/
		Cdr[Cdr[Cdr[m]]] \= NIL
	)
		return wrongargs(n);
	if (\atomic(Car[m])) return error("bad symbol name", Car[m]);
	if (atomic(Car[Cdr[m]])) return badargl(n);
	f := Car[m];
	! Protecting M[NCDR] is not necessary,
	! since it already is protected in EVAL()
	Cdr[f] := alloc(S_lambda, Cdr[m]);
	return f;
end

decl eval(1);

! Evaluate N=(DEFINE M eval[M2])
! The name M already has been added to the
! symbol table by READ().
dodefine(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	if (\atomic(Car[m])) return error("bad symbol name", Car[m]);
	! Bind (and thereby protect) the unevaluated expression
	! Cdr[Car[m]] := Car[Car[m]];
	! Protect the unevaluated expression
	save(Car[Cdr[m]]);
	! And then evaluate it
	Cdr[Car[m]] := eval(Car[Cdr[m]]);
	! Fix recursive references in closures
	Tmp2 := alloc(Car[m], NIL);
	Bstack := alloc(Tmp2, Bstack);
	Tmp2 := NIL;
	fixclosures();
	bunsave(1);
	unsave(1);
	return Car[m];
end

! Evaluate N=(QUOTE M)
doquote(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	return (Car[m]);
end

! Evaluate N=(EVAL M)
doeval(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	return Car[m];
end

! Check if N is an alist.
isalist(n) do
	if (atomic(n)) return 0;
	while (n \= NIL) do
		if (atomic(Car[n]) \/ \atomic(Car[Car[n]]))
			return 0;
		n := Cdr[n];
	end
	return %1;
end

! Check LAMBDA syntax and create closure from lambda expression.
dolambda(n) do var m;
	m := Cdr[n];
	if (	m = NIL \/ Cdr[m] = NIL \/
		Cdr[Cdr[m]] \= NIL /\ Cdr[Cdr[Cdr[m]]] \= NIL
	)
		return wrongargs(n);
	if (Cdr[Cdr[m]] \= NIL /\ \isalist(Car[Cdr[Cdr[m]]]))
		return error("bad environment in 'lambda'",
			Car[Cdr[Cdr[m]]]);
	if (atomic(Car[m])) return badargl(n);
	m := Car[m];
	while (m \= NIL) do
		if (\atomic(Car[m])) return badargl(n);
		if (atomic(Cdr[m])) return error("improper argument list", n);
		m := Cdr[m];
	end
	if (Cdr[Cdr[Cdr[n]]] = NIL) return closure(n);
	return n;
end

! Create a flat copy of a list.
! Store a reference to the last member
! of the copy in lastp.
flatcopy(n, lastp) do var a, m, last;
	if (n = NIL) do
		lastp[0] := NIL;
		return NIL;
	end
	m := alloc(NIL, NIL);
	save(m);
	a := m;
	last := m;
	while (n \= NIL) do
		Car[a] := Car[n];
		last := a;
		n := Cdr[n];
		if (n \= NIL) do
			Cdr[a] := alloc(NIL, NIL);
			a := Cdr[a];
		end
	end
	unsave(1);
	lastp[0] := last;
	return m;
end

! Create a 2-level copy of the Lstack
! and fix cyclic references. N = Lstack.
copy(n) do var m, a, p, last, unused;
	if (n = NIL) return NIL;
	m := alloc(NIL, NIL);
	a := m;
	save(m);
	while (n \= NIL) do
		if (Cdr[n] = NIL \/ Cdr[Cdr[n]] = NIL)
			fatal("internal: unexpected NIL in COPY");
		Car[a] := flatcopy(Car[n], @last);
		Cdr[a] := alloc(NIL, NIL);
		Car[Cdr[a]] := flatcopy(Car[Cdr[n]], @unused);
		Cdr[Cdr[a]] := alloc(last, NIL);
		n := Cdr[Cdr[Cdr[n]]];
	end
	unsave(1);
	return m;
end

! Copy names and values of the symbol table
! into an alist.
copybindings() do var y, p, ny;
	y := Symbols;
	p := alloc(NIL, NIL);
	ny := p;
	save(p);
	while (y \= NIL) do
		Car[p] := alloc(Car[y], Cdr[Car[y]]);
		y := Cdr[y];
		if (y \= NIL) Cdr[p] := alloc(NIL, NIL);
		p := Cdr[p];
	end
	unsave(1);
	return ny;
end

! Restore values of the symbol table.
restorebindings(values) do var b;
	while (values \= NIL) do
		b := Car[values];
		Cdr[Car[b]] := Cdr[b];
		values := Cdr[values];
	end
end

! Capture interpreter state, creating a continuation.
capture() do var cont, x;
	cont := alloc(S_continuation, NIL);
	save(cont);
	Cdr[cont] := alloc(NIL, NIL);
	Car[Cdr[cont]] := alloc(Car[Mstack], NIL);
	Cdr[Cdr[cont]] := alloc(NIL, NIL);
	Car[Cdr[Cdr[cont]]] := copy(Lstack);
	Cdr[Cdr[Cdr[cont]]] := alloc(Cdr[Cdr[Stack]], NIL);
	Cdr[Cdr[Cdr[Cdr[cont]]]] := alloc(NIL, NIL);
	Car[Cdr[Cdr[Cdr[Cdr[cont]]]]] := flatcopy(Bstack, @x);
	Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]] := alloc(Estack, NIL);
	Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]]] := alloc(NIL, NIL);
	Car[Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]]]] := copybindings();
	unsave(1);
	return cont;
end

! Check the arguments of a continuation.
ckcont(n) do
	if (	Cdr[n] = NIL \/ Cdr[Cdr[n]] = NIL \/
		Cdr[Cdr[Cdr[n]]] = NIL \/ Cdr[Cdr[Cdr[Cdr[n]]]] = NIL \/
		Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]] = NIL \/
		Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]]] = NIL \/
		Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]]]] \= NIL
	)
		error("invalid continuation", n);
	return n;
end

! Re-activate a captured continuation N.
resume(n) do var m, state, q, x;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	ckcont(Car[n]);
	if (Errflag) return NIL;
	state := Cdr[Car[n]];
	Car[Mstack] := Car[Car[state]];
	Lstack := Car[Cdr[state]];
	Stack := Car[Cdr[Cdr[state]]];
	Bstack := flatcopy(Car[Cdr[Cdr[Cdr[state]]]], @x);
	Estack := Car[Cdr[Cdr[Cdr[Cdr[state]]]]];
	restorebindings(Car[Cdr[Cdr[Cdr[Cdr[Cdr[state]]]]]]);
	! Quote argument to prevent double evaluation.
	q := alloc(S_quote, NIL);
	save(q);
	Cdr[q] := alloc(Car[m], NIL);
	unsave(1);
	return q;
end

! Evaluate N=(CALL/CC M)
docallcc(n) do var m;
	if (Evlev > 1) do
		error("'call/cc' cannot be used here; this is a known bug",
		%1);
		return NIL;
	end
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := alloc(Car[m], NIL);
	save(m);
	Cdr[m] := alloc(capture(), NIL);
	unsave(1);
	return m;
end

! Check whether (CAR NP[0]) is a builtin symbol.
! If it is one, run the appropriate routine, save
! its result in NP[0], and return %1.
! Return 0 if (CAR NP[0]) is not a builtin procedure.
builtin(np) do var n, y;
	n := np[0];
	y := Car[n];
	ie (Errflag) return 0;
	else ie (y = S_car) n := docar(n);
	else ie (y = S_cdr) n := docdr(n);
	else ie (y = S_cons) n := docons(n);
	else ie (y = S_null) n := donull(n);
	else ie (y = S_eq) n := doeq(n);
	else ie (y = S_atom) n := doatom(n);
	else ie (y = S_quote) n := doquote(n);
	else ie (y = S_lambda) n := dolambda(n);
	else ie (y = S_continuation) n := ckcont(n);
	else ie (y = S_list) n := dolist(n);
	else ie (y = S_explode) n := doexplode(n);
	else ie (y = S_implode) n := doimplode(n);
	else ie (y = S_define) n := dodefine(n);
	else ie (y = S_defun) n := dodefun(n);
	else return 0;
	np[0] := n;
	return %1;
end

! Bind the arguments of a LAMBDA function.
! For a lambda application N=((LAMBDA (X1 ... Xn) S [ENV]) Y1 ... Yn)
! this includes the following steps for j in {1,...,n}:
!	1) save Xj in V
!	2) save the value of Xj
!	3) bind Xj to Yj
! This routine results in  S' = S[X1/Y1] ... [Xn/Yn].
! S->S' is performed by creating a new context. BINDARGS()
! has no function result. It is used by EVAL() to perform
! BETA reduction steps.
bindargs(n, name) do
	var	fa,	! Formal arg list
		aa,	! Actual arg list
		e;	! S, as above
	var	env;	! Optional lexical environment
	var	p;
	var	na;

	if (Errflag) return NIL;
	fa := Car[Cdr[Car[n]]];
	aa := Cdr[n];
	p := Cdr[Cdr[Car[n]]];
	e := Car[p];
	env := Cdr[p] \= NIL -> Car[Cdr[p]]: NIL;
	bsave(NIL);
	na := 1;
	while (fa \= NIL /\ aa \= NIL) do
		Car[Bstack] := alloc(Car[fa], Car[Bstack]);	! Save name
		save(Cdr[Car[fa]]);			! Save value
		Cdr[Car[fa]] := Car[aa];		! Bind arg
		fa := Cdr[fa];
		aa := Cdr[aa];
		na := na+1;
	end
	if (Reclflag) fixclosures();
	while (env \= NIL) do
		p := Car[env];
		Car[Bstack] := alloc(Car[p], Car[Bstack]);! Save name
		save(Cdr[Car[p]]);		! Save value
		Cdr[Car[p]] := Cdr[p];		! Bind lex val
		env := Cdr[env];
		na := na+1;
	end
	if (Statflag) Reductions.count(na);
	ie (fa \= NIL \/ aa \= NIL) do
		wrongargs(n);
		n := NIL;
	end
	else do
		n := e;
	end
	if (Ftflag) do
		save(Function);
		Function := name;
		save(Frame);
		Frame := Stack;
	end
	return n;
end

! Print application of traced function N in the form
!	* (NAME A1 ... An)
! PRINT() cannot be used because it would print NAME in
! its expanded (LAMBDA...) form which is not desirable.
printtrace(n) do
	Prefix := "* ";
	pr("(");
	Quoted := 1;
	_print(Trace);
	while (1) do
		n := Cdr[n];
		if (n = NIL) leave;
		pr("\s");
		_print(Car[n]);
	end
	pr(")"); nl();
	Prefix := 0;
end

! Print depth of stack N.
prdepth(n) do var k;
	k := 0;
	while (n \= NIL) do
		n := Cdr[n];
		k := k+1;
	end
	prnum(k, 6); pr("\s");
end

! Debugging: dump interpreter state
dumpstate(s, m) do
	pr(s); nl();
	pr("\s\s\s\s\s\s\s"); pr("Mode        = "); prnum(m-'0',0); nl();
	prdepth(Lstack);      pr("Lstack      = "); print(Lstack); nl();
	pr("\s\s\s\s\s\s\s"); pr("Mstack      = "); print(Mstack); nl();
	prdepth(Stack);       pr("Car[Stack]  = "); print(Car[Stack]); nl();
	pr("\s\s\s\s\s\s\s"); pr("Stack0      = "); prnum(Stack0,0); nl();
	prdepth(Bstack);      pr("Bstack      = "); print(Bstack); nl();
end

! Evaluate the term N and return its normal form.
! This is the heart of the interpreter:
! An iterative EVAL function with tail-call optimization and call/cc.
! For details on the reduction rules, refer to the manual.
! Basically, EVAL() performs these steps:
!	- If N is NIL, return NIL
!	- If N is a symbol, return its binding
!	- If N is a list (A1 ... An)
!		- If A1 not in {COND,DEFUN,DEFINE,LABEL,LAMBDA,QUOTE}
!			- Create a new list (A1' ... An')
!				where Aj'=EVAL(Aj) for j in {1,...,n}
!		- If A1' (or A1) is a primitive (builtin fn), apply it,
!			resulting in N'
!		- If A1 is a lambda function, BETA-reduce it,
!			resulting in N'
!	- Return N'
!
eval(n) do
	var	m,	! Result node
		m2,	! Root of result lists
		a;	! Used to append to result
	! Evaluator states
	const	MATOM = '0',	! Processing Atom
		MLIST = '1',	! Processing List
		MBETA = '2',	! Beta-reducing
		MBIND = '3',	! Processing bindings of LABEL
		MCOND = '4';	! Processing predicates of COND
	var	mode,	! Current state
		cf;	! Continuation flag
	var	nm;	! name of function to apply

	Evlev := Evlev + 1;
	save(Stack0);
	Stack0 := Stack;
	mode := MATOM;
	cf := 0;
	while (\Errflag) do
		ie (n = NIL) do			! () -> ()
			m := NIL;
		end
		else ie (atomic(n)) do		! Symbol -> Value
			if (Statflag) Reductions.count(1);
			m := Cdr[n] = Car[n]-> n: Cdr[n];
		end
		else do				! List (...) and Pair (X.Y)
			! This block is used to DESCEND into lists.
			! The following nodes/variables will be saved:
			!	1) the original list (on Stack)
			!	2) the current state (on Mstack)
			!	3) the root of the result list (on Lstack)
			!	4) a ptr to the next free node
			!		in the result list (on Lstack)
			!	5) a ptr to the next member of
			!		the original list (on Lstack)
			m := Car[n];
			if (atomic(Cdr[n])) do
				error("improper list in application", n);
				n := NIL;
			end
			save(n);	! Save original list
			msave(mode);
			! These builtins are call-by-name
			ie (	m = S_cond \/ m = S_define \/ m = S_defun \/
				m = S_label \/ m = S_quote \/ m = S_lambda \/
				m = S_continuation
			) do
				lsave(NIL);
				lsave(NIL);
				lsave(n);	! Root of result list
				n := NIL;
			end
			else do
				a := alloc(NIL, NIL);
				lsave(a);
				lsave(Cdr[n]);
				lsave(a);	! Root of result list
				n := Car[n];
			end
			mode := MLIST;
			loop;
		end
		! The following loop is used to ASCEND back to the
		! root of a list, thereby performing BETA reduction
		! and creating result lists.
		while (1) ie (mode = MBETA) do
			! Finish BETA reduction
			unbindargs();
			unsave(1);	! Original list
			mode := munsave();
		end
		else ie (mode = MLIST) do	! Append to list, reduce
			n := Car[Cdr[Lstack]];	! Next member
			a := Car[Cdr[Cdr[Lstack]]];	! Place to append to
			m2 := lunsave(1);	! Root of result list
			if (a \= NIL)		! Append new member
				Car[a] := m;
			ie (n = NIL) do		! End of list
				m := m2;
				lunsave(2);	! Drop N,A
 				! Drop original list, remember first element
				nm := Car[unsave(1)];
				save(m);	! Save result
				ie (builtin(@m))	! Builtin?
					;
				else ie (Car[m] = S_cond) do
					! COND is special
					n := setupcond(m);
					cf := 2;
					mode := MCOND;
				end
				else ie (\atomic(Car[m]) /\
					Car[m] \= NIL /\
					Car[Car[m]] = S_lambda
				) do
					! Application: reduce ((lambda...)...)
					if (Trace = nm) printtrace(m);
					nm := atomic(nm)-> nm: NIL;
					! If the caller is also BETA-reducing,
					! this is a TAIL application.
					if (Car[Car[Mstack]] & ~AFLAG = MBETA)
					do
						! In this case, remove the
						! caller's context.
						Tmp2 := unsave(1); ! M
						unbindargs();
						unsave(1);
						munsave();
						save(Tmp2);
						Tmp2 := NIL;
					end
					bindargs(m, nm);
					! N=S in ((LAMBDA (...) S) ...)
					n := Car[Cdr[Cdr[Car[m]]]];
					cf := 2;
					mode := MBETA;
				end
				else ie (Car[m] = S_label) do
					! LABEL is special
					ie (setuplabel(m) \= NIL)
						n := evallabel();
					else
						n := NIL;
					cf := 2;
					mode := MBIND;
				end
				else ie (Car[m] = S_callcc) do
					! CALL/CC is special
! dumpstate("call/cc", mode);
					n := docallcc(m);
					cf := 1;
				end
				else ie (\atomic(Car[m]) /\
					Car[m] \= NIL /\
					Car[Car[m]] = S_continuation
				) do
					n := resume(m);
! dumpstate("resume", mode);
					cf := 1;
				end
				else ie (Car[m] = S_eval) do
					! EVAL is special
					n := doeval(m);
					cf := 1;
				end
				else do
					if (Car[m] = S_bottom) error("", m);
					error("application of non-function",
						nm);
					n := NIL;
				end
				if (cf \= 2) do
					unsave(1);	! Drop result
					mode := munsave();
				end
				! Continuation requested.
				! Leave the ASCENDING loop and descend
				! into the continuation N.
				if (cf) leave;
			end
			else do			! Append to list
				lsave(m2);
				! Create space for next argument
				Cdr[a] := alloc(NIL, NIL);
				Car[Cdr[Cdr[Lstack]]] := Cdr[a];
				Car[Cdr[Lstack]] := Cdr[n];
				n := Car[n];	! Evaluate next member
				leave;
			end
		end
		else ie (mode = MCOND) do
			n := evalclause(m);
			if (Car[Bstack] = NIL) do
				unsave(1);	! Original list
				bunsave(1);
				mode := munsave();
			end
			cf := 1;
			leave;
		end
		else ie (mode = MBIND) do
			ie (nextlabel(m) = NIL) do
				n := finishlabel(
					Car[Car[Mstack]] & ~AFLAG = MBETA);
				mode := MBETA;
			end
			else do
				n := evallabel();
			end
			cf := 1;
			leave;
		end
		else do	! Atom
			leave;
		end
		if (cf) do	! Do a continuation, if requested
			cf := 0;
			loop;
		end
		if (Stack = Stack0) leave;
	end
	while (Stack \= Stack0) unsave(1);
	Stack0 := unsave(1);
	Evlev := Evlev - 1;
	return m;		! Return the evaluated expr
end

! Test N for being a list of single-character symbols
! and print such lists in imploded form.
printxsym(n) do var i, s::TEXTLEN+2, c;
	s::0 := '#';
	i := 0;
	while (1) do
		if (i >= TEXTLEN-1) return 0;
		if (n = NIL) leave;
		if (atomic(n) \/ \atomic(Car[n])) return 0;
		c := Car[Car[Car[n]]] & 255;
		if (Cdr[Car[Car[n]]] \= NIL) return 0;
		n := Cdr[n];
		i := i+1;
		s::i := c;
	end
	i := i+1;
	s::i := 0;
	pr(s);
	return 1;
end

! Print expressions of the form (QUOTE X) as 'X.
printquote(n) do
	if (	Car[n] = S_quote /\
		Cdr[n] \= NIL /\
		Cdr[Cdr[n]] = NIL
	) do
		pr("'");
		_print(Car[Cdr[n]]);
		return 1;
	end
	return 0;
end

! Print a closure. This may be a cyclic structure, so
! suppress the third agument, if any.
printclosure(n) do
	if (Prclflag) return;
	if (	Car[n] = S_lambda /\
		Cdr[n] \= NIL /\
		Cdr[Cdr[n]] \= NIL /\
		Cdr[Cdr[Cdr[n]]] \= NIL
	) do
		pr("(lambda ");
		_print(Car[Cdr[n]]);
		pr("\s");
		_print(Car[Cdr[Cdr[n]]]);
		pr(" ...)");
		return %1;
	end
	return 0;
end

! Recursively print the term N.
_print(n) do
	ie (n = NIL) do
		pr("()");
	end
	else ie (Car[n] & AFLAG) do
		! Characters are limited to the symbol table
		error("unprintable object", %1);
	end
	else ie (atomic(n)) do var s::TEXTLEN, i;
		if (\Quoted) do		! Symbol
			pr("'");
			Quoted := 1;
		end
		i := 0;
		n := Car[n];
		while (n \= NIL) do
			s::i := Car[n];
			if (i < TEXTLEN-2) i := i+1;
			n := Cdr[n];
		end
		s::i := 0;
		pr(s);
	end
	else do				! List
		if (\Quoted) do
			pr("'");
			Quoted := 1;
		end
		if (Car[n] = S_continuation) do
			pr("(continuation ...)");
			return;
		end
		if (printclosure(n))
			return;
		if (Implflag /\ printxsym(n)) return;
		if (printquote(n)) return;
		pr("(");
		while (n \= NIL) do
			_print(Car[n]);
			n := Cdr[n];
			if (atomic(n)) do
				pr(".");
				_print(n);
				n := NIL;
			end
			if (n \= NIL) pr("\s");
		end
		pr(")");
	end
end

! Print quoted expression.
print(n) do
	Quoted := 0;
	_print(n);
end

! Reset interpreter state.
resetstate() do
	Stack := NIL;
	Lstack := NIL;
	Bstack := NIL;
	Estack := NIL;
	Frame := NIL;
	Function := NIL;
	Evlev := 0;
end

! Initialize interpreter variables.
init1() do var i, n;
	RELEASE := "2005-04-29";
	! Misc. variables
	NIL := POOLSIZE;
	Level := 0;
	resetstate();
	Symbols := NIL;
	Safesymbols := NIL;
	Tmp := NIL;
	Tmp2 := NIL;
	Loadlev := 0;
	Trace := NIL;
	Pp := 0;
	t.newline(Nlseq);
	Nll := str.length(Nlseq);
	Wrapcol := 0;
	Prefix := 0;
	Maxuse := 0;
	Lastuse := 0;
	Statflag := 0;
	Ftflag := 1;
	Ntrace := 10;
	Implflag := 1;
	Prclflag := 0;
	Reclflag := 0;
	Image::0 := 0;
	Line := 1;
	! Initialize Freelist
	Free := NIL;
	! Clear input buffer
	Infile := 0;
	Buffer := Bufvec;
	Input := T3X.SYSIN;
	Bp := 0;
	Ep := 0;
	Rejected := EOF;
end

! Second stage of initialization:
! protect registers from GC,
! build the free list,
! create builtin symbols.
init2() do
	! Protect base registers
	Base[0] := @Symbols;
	Base[1] := @Stack;
	Base[2] := @Mstack;
	Base[3] := @Lstack;
	Base[4] := @Bstack;
	Base[5] := @Estack;
	Base[6] := @Tmp;
	Base[7] := @Tmp2;
	Base[8] := @Safesymbols;
	! Create builtin symbols
	S_symbols := NIL;
	S_nil := addsym("nil", NIL);	! First GC will be triggered HERE
	S_atom := addsym("atom", 0);
	S_bottom := addsym("bottom", 0);
	S_callcc := addsym("call/cc", 0);
	S_car := addsym("car", 0);
	S_cdr := addsym("cdr", 0);
	S_cond := addsym("cond", 0);
	S_cons := addsym("cons", 0);
	S_continuation := addsym("continuation", 0);
	S_define := addsym("define", 0);
	S_defun := addsym("defun", 0);
	S_eq := addsym("eq", 0);
	S_eval := addsym("eval", 0);
	S_explode := addsym("explode", 0);
	S_false := addsym("*f*", 0);
	S_implode := addsym("implode", 0);
	S_label := addsym("label", 0);
	S_lambda := addsym("lambda", 0);
	S_list := addsym("list", 0);
	S_null := addsym("null", 0);
	S_quote := addsym("quote", 0);
	S_symbols := addsym("*symbols*", 0);
	S_true := addsym("*t*", 0);
	S_t := addsym("t", S_true);
	S_last := addsym("**", 0);
	Mstack := alloc(NIL, NIL);
end

! Reset the reduction counter
clearstats() do
	Reductions.reset();
	Allocations.reset();
end

! Print the number of reductions done in last EVAL
printstats() do
	Reductions.printvalue("reductions");
	Allocations.printvalue("nodes allocated");
end

! This is the top-level Read-Eval-Print loop.
toploop() do var n, def;
	while (1) do
		Errflag := 0;
		n := read();
		if (n = EOF) leave;
		save(n);
		Safesymbols := copybindings();
		if (Statflag) clearstats();
		n := eval(unsave(1));
		ie (\Errflag) do
			Cdr[S_last] := n;
			if (\Loadlev) do
				if (\O_batch) pr("=> ");
				print(n); nl();
			end
			if (Stack \= NIL)
				fatal("internal: unbalanced stack");
			if (Statflag /\ \Loadlev) printstats();
		end
		else do
			restorebindings(Safesymbols);
		end
		resetstate();
		while (Car[Mstack] \= NIL) munsave();
	end
end

! Load initial node pool image
loadimage(p) do var fd, n, v, i, buf::6;
	fd := t.open(p, T3X.OREAD);
	if (fd < 0) do
		error("cannot open image", %1);
		pr(">>> "); pr(p); nl();
		halt 1;
	end
	t.read(fd, buf, 7);
	t.read(fd, @n, t.bpw());
	if (	t.memcomp(buf, "ALISP", 5) \/
		buf::5 \= t.bpw() \/
		buf::6 \= MAJORVERSION \/
		n \= POOLSIZE
	) do
		error("bad image", %1);
		pr(">>> "); pr(p); nl();
		halt 1;
	end
	v := imagevars();
	i := 0;
	while (v[i]) do
		t.read(fd, v[i], t.bpw());
		i := i+1;
	end
	if (	t.read(fd, Car, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw() \/
		t.read(fd, Cdr, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw()
	) do
		error("bad image", %1);
		pr(">>> "); pr(p); nl();
		halt 1;
	end
	t.close(fd);
	if (\O_batch) do
		pr("Image loaded"); nl();
	end
end

! Print usage and fail.
usage() do
	error("Usage: alisp [-b] [image]", %1);
	halt 1;
end

! Evaluate the command line options
getopts() do var a::TEXTLEN, i, j, k;
	O_batch := 0;
	i := 1;
	while (1) do
		k := t.getarg(i, a, TEXTLEN);
		if (k < 1 \/ a::0 \= '-') leave;
		for (j=1, k)
			ie (a::j = 'b' \/ a::j = 'B') O_batch := 1;
			else usage();
		i := i+1;
	end
	if (k > 0) str.copy(Image, a);
end

do
	init1();
	getopts();
	init2();
	if (\O_batch) do
		pr("ArrowLISP ");
		pr(RELEASE);
		pr(" (C) 1998-2005 Nils M Holm"); nl();
	end
	if (DEBUG) do
		pr("Debugging enabled"); nl();
	end
	if (\O_batch) do
		pr("Poolsize="); prnum(POOLSIZE, 0); nl();
	end
	if (Image::0) loadimage(Image);
	toploop();
	bye();
end
